@@ -63,7 +63,8 @@ let checkForChangesToSourceLocal
         && not (Fpcache.excelFile pathFrom)
         && match archStamp with
              Some (Fileinfo.InodeStamp inode) -> sourceInfo.Fileinfo.inode = inode
-           | Some (Fileinfo.CtimeStamp ctime) -> true
+           | Some (Fileinfo.NoStamp)          -> true
+           | Some (Fileinfo.RescanStamp)      -> false
            | None                             -> false   in
       let ressClearlyUnchanged =
            not clearlyChanged
@@ -101,11 +102,38 @@ let checkForChangesToSourceLocal
            Transfer aborted."
           (Fspath.toPrintString (Fspath.concat fspathFrom pathFrom))))
 
+let mcheckForChangesToSource =
+  Umarshal.(prod2
+              (prod4 Path.mlocal Props.m Os.mfullfingerprint (option Fileinfo.mstamp) id id)
+              (prod3 Osx.mressStamp (option Os.mfullfingerprint) bool id id)
+              id id)
+
+let archStamp_to_compat251 = function
+  | Some stamp -> Some (Fileinfo.stamp_to_compat251 stamp)
+  | None -> None
+
+let archStamp_of_compat251 = function
+  | Some stamp -> Some (Fileinfo.stamp_of_compat251 stamp)
+  | None -> None
+
+let convV0 = Remote.makeConvV0FunArg
+  (fun (fspathFrom,
+          ((pathFrom, archDesc, archFp, archStamp), (archRess, newFpOpt, paranoid))) ->
+       (fspathFrom,
+          (pathFrom, Props.to_compat251 archDesc, archFp,
+            archStamp_to_compat251 archStamp, archRess, newFpOpt, paranoid)))
+  (fun (fspathFrom,
+          (pathFrom, archDesc, archFp, archStamp, archRess, newFpOpt, paranoid)) ->
+       (fspathFrom,
+          ((pathFrom, Props.of_compat251 archDesc, archFp,
+            archStamp_of_compat251 archStamp), (archRess, newFpOpt, paranoid))))
+
 let checkForChangesToSourceOnRoot =
   Remote.registerRootCmd
-    "checkForChangesToSource"
+    "checkForChangesToSource" ~convV0
+    mcheckForChangesToSource Umarshal.unit
     (fun (fspathFrom,
-          (pathFrom, archDesc, archFp, archStamp, archRess, newFpOpt, paranoid)) ->
+          ((pathFrom, archDesc, archFp, archStamp), (archRess, newFpOpt, paranoid))) ->
       checkForChangesToSourceLocal
         fspathFrom pathFrom archDesc archFp archStamp archRess newFpOpt paranoid;
       Lwt.return ())
@@ -113,7 +141,7 @@ let checkForChangesToSourceOnRoot =
 let checkForChangesToSource
       root pathFrom archDesc archFp archStamp archRess newFpOpt paranoid =
   checkForChangesToSourceOnRoot
-    root (pathFrom, archDesc, archFp, archStamp, archRess, newFpOpt, paranoid)
+    root ((pathFrom, archDesc, archFp, archStamp), (archRess, newFpOpt, paranoid))
 
 (****)
 
@@ -146,6 +174,8 @@ let rec fingerprintPrefix fspath path offset len accu =
 let fingerprintPrefixRemotely =
   Remote.registerServerCmd
     "fingerprintSubfile"
+    Umarshal.(prod3 Fspath.m Path.mlocal Uutil.Filesize.m id id)
+    Umarshal.(list Fingerprint.m)
     (fun _ (fspath, path, len) ->
        Lwt.return (fingerprintPrefix fspath path 0L len []))
 
@@ -168,11 +198,48 @@ let validFilePrefix connFrom fspathFrom pathFrom fspathTo pathTo info desc =
   end else
     Lwt.return None
 
+(* IMPORTANT!
+   This is the 2.51-compatible version of type [transferStatus]. It must always
+   remain exactly the same as the type [transferStatus] in version 2.51.5. This
+   means that if any of the types it is composed of changes then for each
+   changed type also a 2.51-compatible version must be created. *)
+type transferStatus251 =
+    TransferSucceeded of Fileinfo.t251
+  | TransferNeedsDoubleCheckAgainstCurrentSource of Fileinfo.t251 * Os.fullfingerprint
+  | TransferFailed of string
+
 type transferStatus =
     TransferSucceeded of Fileinfo.t
   | TransferNeedsDoubleCheckAgainstCurrentSource of Fileinfo.t * Os.fullfingerprint
   | TransferFailed of string
 
+let mtransferStatus = Umarshal.(sum3
+                                  Fileinfo.m
+                                  (prod2 Fileinfo.m Os.mfullfingerprint id id)
+                                  string
+                                  (function
+                                   | TransferSucceeded a -> I31 a
+                                   | TransferNeedsDoubleCheckAgainstCurrentSource (a, b) -> I32 (a, b)
+                                   | TransferFailed a -> I33 a)
+                                  (function
+                                   | I31 a -> TransferSucceeded a
+                                   | I32 (a, b) -> TransferNeedsDoubleCheckAgainstCurrentSource (a, b)
+                                   | I33 a -> TransferFailed a))
+
+let transferStatus_to_compat251 (st : transferStatus) : transferStatus251 =
+  match st with
+  | TransferSucceeded info -> TransferSucceeded (Fileinfo.to_compat251 info)
+  | TransferNeedsDoubleCheckAgainstCurrentSource (info, fp) ->
+      TransferNeedsDoubleCheckAgainstCurrentSource (Fileinfo.to_compat251 info, fp)
+  | TransferFailed s -> TransferFailed s
+
+let transferStatus_of_compat251 (st : transferStatus251) : transferStatus =
+  match st with
+  | TransferSucceeded info -> TransferSucceeded (Fileinfo.of_compat251 info)
+  | TransferNeedsDoubleCheckAgainstCurrentSource (info, fp) ->
+      TransferNeedsDoubleCheckAgainstCurrentSource (Fileinfo.of_compat251 info, fp)
+  | TransferFailed s -> TransferFailed s
+
 (* Paranoid check: recompute the transferred file's fingerprint to match it
    with the archive's.  If the old
    fingerprint was a pseudo-fingerprint, we can't tell just from looking at the
@@ -217,8 +284,12 @@ let saveTempFileLocal (fspathTo, (pathTo, realPathTo, reason)) =
         reason
         (Fspath.toDebugString (Fspath.concat fspathTo savepath))))
 
+let convV0 = Remote.makeConvV0FunRet Fileinfo.to_compat251 Fileinfo.of_compat251
+
 let saveTempFileOnRoot =
-  Remote.registerRootCmd "saveTempFile" saveTempFileLocal
+  Remote.registerRootCmd "saveTempFile" ~convV0
+    Umarshal.(prod3 Path.mlocal Path.mlocal string id id) Fileinfo.m
+    saveTempFileLocal
 
 (****)
 
@@ -408,10 +479,10 @@ let processTransferInstruction conn (file_id, ti) =
        ignore (Remote.MsgIdMap.find file_id !decompressor ti))
 
 let marshalTransferInstruction =
-  (fun (file_id, (data, pos, len)) rem ->
+  (fun _ (file_id, (data, pos, len)) rem ->
      (Remote.encodeInt file_id :: (data, pos, len) :: rem,
       len + Remote.intSize)),
-  (fun buf pos ->
+  (fun _ buf pos ->
      let len = Bytearray.length buf - pos - Remote.intSize in
      (Remote.decodeInt buf pos, (buf, pos + Remote.intSize, len)))
 
@@ -426,7 +497,7 @@ let showPrefixProgress id kind =
   | _                -> ()
 
 let compress conn
-     (biOpt, fspathFrom, pathFrom, fileKind, sizeFrom, id, file_id) =
+     ((biOpt, fspathFrom, pathFrom, fileKind), (sizeFrom, id, file_id)) =
   Lwt.catch
     (fun () ->
        streamTransferInstruction conn
@@ -461,7 +532,29 @@ let compress conn
        Util.convertUnixErrorsToTransient "transferring file contents"
          (fun () -> raise e))
 
-let compressRemotely = Remote.registerServerCmd "compress" compress
+let mdata = Umarshal.(sum3 unit Uutil.Filesize.m unit
+                        (function
+                         | `DATA -> I31 ()
+                         | `DATA_APPEND a -> I32 a
+                         | `RESS -> I33 ())
+                        (function
+                         | I31 () -> `DATA
+                         | I32 a -> `DATA_APPEND a
+                         | I33 () -> `RESS))
+
+let mcompress = Umarshal.(prod2
+                            (prod4 (option Transfer.Rsync.mrsync_block_info) Fspath.m Path.mlocal mdata id id)
+                            (prod3 Uutil.Filesize.m Uutil.File.m int id id)
+                            id id)
+
+let convV0 = Remote.makeConvV0FunArg
+  (fun ((biOpt, fspathFrom, pathFrom, fileKind), (sizeFrom, id, file_id)) ->
+       (biOpt, fspathFrom, pathFrom, fileKind, sizeFrom, id, file_id))
+  (fun (biOpt, fspathFrom, pathFrom, fileKind, sizeFrom, id, file_id) ->
+       ((biOpt, fspathFrom, pathFrom, fileKind), (sizeFrom, id, file_id)))
+
+let compressRemotely =
+  Remote.registerServerCmd "compress" ~convV0 mcompress Umarshal.unit compress
 
 let close_all infd outfd =
   Util.convertUnixErrorsToTransient
@@ -583,7 +676,7 @@ let transferFileContents
          debug (fun () -> Util.msg "Starting the actual transfer\n");
          decompressor := Remote.MsgIdMap.add file_id decompr !decompressor;
          compressRemotely connFrom
-           (bi, fspathFrom, pathFrom, fileKind, srcFileSize, id, file_id)
+           ((bi, fspathFrom, pathFrom, fileKind), (srcFileSize, id, file_id))
            >>= fun () ->
          decompressor :=
            Remote.MsgIdMap.remove file_id !decompressor; (* For GC *)
@@ -789,8 +882,8 @@ let prepareExternalTransfer fspathTo pathTo =
       false
 
 let finishExternalTransferLocal connFrom
-      (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
-       update, desc, fp, ress, id) =
+      ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+       (update, desc, fp, ress, id)) =
   let info = Fileinfo.get false fspathTo pathTo in
   if
     info.Fileinfo.typ <> `FILE ||
@@ -805,9 +898,35 @@ let finishExternalTransferLocal connFrom
   Xferhint.insertEntry fspathTo pathTo fp;
   Lwt.return res
 
+let convV0 = Remote.makeConvV0Funs
+  (fun ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+         (update, desc, fp, ress, id)) ->
+       (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
+         update, Props.to_compat251 desc, fp, ress, id))
+  (fun (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
+         update, desc, fp, ress, id) ->
+       ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+         (update, Props.of_compat251 desc, fp, ress, id)))
+  transferStatus_to_compat251
+  transferStatus_of_compat251
+
+let mcopyOrUpdate = Umarshal.(sum2 unit (prod2 Uutil.Filesize.m Uutil.Filesize.m id id)
+                                (function
+                                 | `Copy -> I21 ()
+                                 | `Update (a, b) -> I22 (a, b))
+                                (function
+                                 | I21 () -> `Copy
+                                 | I22 (a, b) -> `Update (a, b)))
+
+let mfinishExternalTransfer = Umarshal.(prod2
+                                          (prod5 Fspath.m Path.mlocal Fspath.m Path.mlocal Path.mlocal id id)
+                                          (prod5 mcopyOrUpdate Props.m Os.mfullfingerprint Osx.mressStamp Uutil.File.m id id)
+                                          id id)
+
 let finishExternalTransferOnRoot =
   Remote.registerRootCmdWithConnection
-    "finishExternalTransfer" finishExternalTransferLocal
+    "finishExternalTransfer" ~convV0
+    mfinishExternalTransfer mtransferStatus finishExternalTransferLocal
 
 let copyprogReg = Lwt_util.make_region 1
 
@@ -850,14 +969,14 @@ let transferFileUsingExternalCopyprog
              l (if l="" then "" else "\n"));
   Uutil.showProgress id (Props.length desc) "ext";
   finishExternalTransferOnRoot rootTo rootFrom
-    (snd rootFrom, pathFrom, fspathTo, pathTo, realPathTo,
-     update, desc, fp, ress, id)
+    ((snd rootFrom, pathFrom, fspathTo, pathTo, realPathTo),
+     (update, desc, fp, ress, id))
 
 (****)
 
 let transferFileLocal connFrom
-      (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
-       update, desc, fp, ress, id) =
+      ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+       (update, desc, fp, ress, id)) =
   let (tempInfo, isTransferred) =
     fileIsTransferred fspathTo pathTo desc fp ress in
   if isTransferred then begin
@@ -894,8 +1013,33 @@ let transferFileLocal connFrom
                Lwt.return (`DONE (status, None))
              end)
 
+let convV0 = Remote.makeConvV0Funs
+  (fun ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+         (update, desc, fp, ress, id)) ->
+       (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
+         update, Props.to_compat251 desc, fp, ress, id))
+  (fun (fspathFrom, pathFrom, fspathTo, pathTo, realPathTo,
+         update, desc, fp, ress, id) ->
+       ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
+         (update, Props.of_compat251 desc, fp, ress, id)))
+  (function
+   | `DONE (a, b) -> `DONE (transferStatus_to_compat251 a, b)
+   | `EXTERNAL a -> `EXTERNAL a)
+  (function
+   | `DONE (a, b) -> `DONE (transferStatus_of_compat251 a, b)
+   | `EXTERNAL a -> `EXTERNAL a)
+
+let mtransferFile = Umarshal.(sum2 (prod2 mtransferStatus (option string) id id) bool
+                                (function
+                                 | `DONE (a, b) -> I21 (a, b)
+                                 | `EXTERNAL a -> I22 a)
+                                (function
+                                 | I21 (a, b) -> `DONE (a, b)
+                                 | I22 a -> `EXTERNAL a))
+
 let transferFileOnRoot =
-  Remote.registerRootCmdWithConnection "transferFile" transferFileLocal
+  Remote.registerRootCmdWithConnection "transferFile" ~convV0
+    mfinishExternalTransfer mtransferFile transferFileLocal
 
 (* We limit the size of the output buffers to about 512 KB
    (we cannot go above the limit below plus 64) *)
@@ -914,8 +1058,8 @@ let transferFile
   let f () =
     Abort.check id;
     transferFileOnRoot rootTo rootFrom
-      (snd rootFrom, pathFrom, fspathTo, pathTo, realPathTo,
-       update, desc, fp, ress, id) >>= fun status ->
+      ((snd rootFrom, pathFrom, fspathTo, pathTo, realPathTo),
+       (update, desc, fp, ress, id)) >>= fun status ->
     match status with
       `DONE (status, msg) ->
          begin match msg with
