@@ -20,6 +20,7 @@ let debug = Util.debug "props"
 
 module type S = sig
   type t
+  val m : t Umarshal.t
   val dummy : t
   val hash : t -> int -> int
   val similar : t -> t -> bool
@@ -66,6 +67,8 @@ end = struct
 (* "UPDATE.ARCHIVEFORMAT"                                                    *)
 type t = int * int
 
+let m = Umarshal.(prod2 int int id id)
+
 (* This allows us to export NullPerm while keeping the type perm abstract    *)
 let dummy = (0, 0)
 
@@ -298,6 +301,16 @@ type t =
   | IdNamed of string
   | IdNumeric of int
 
+let m = Umarshal.(sum3 unit string int
+                    (function
+                     | IdIgnored -> I31 ()
+                     | IdNamed a -> I32 a
+                     | IdNumeric a -> I33 a)
+                    (function
+                     | I31 () -> IdIgnored
+                     | I32 a -> IdNamed a
+                     | I33 a -> IdNumeric a))
+
 let dummy = IdIgnored
 
 let hash id h =
@@ -441,6 +454,14 @@ let sync =
 
 type t = Synced of float | NotSynced of float
 
+let m = Umarshal.(sum2 float float
+                    (function
+                     | Synced a -> I21 a
+                     | NotSynced a -> I22 a)
+                    (function
+                     | I21 a -> Synced a
+                     | I22 a -> NotSynced a))
+
 let dummy = NotSynced 0.
 
 let extract t = match t with Synced v -> v | NotSynced v -> v
@@ -615,6 +636,8 @@ module TypeCreator : S = struct
 
 type t = string option
 
+let m = Umarshal.(option string)
+
 let dummy = None
 
 let hash t h = Uutil.hash2 (Uutil.hash t) h
@@ -665,6 +688,19 @@ end
 (*                           Properties                                      *)
 (* ------------------------------------------------------------------------- *)
 
+(* IMPORTANT!
+   This is the 2.51-compatible version of type [Props.t]. It must always remain
+   exactly the same as the type [Props.t] in version 2.51.5. This means that if
+   any of the types it is composed of changes then for each changed type also a
+   2.51-compatible version must be created. *)
+type t251 =
+  { perm : Perm.t;
+    uid : Uid.t;
+    gid : Gid.t;
+    time : Time.t;
+    typeCreator : TypeCreator.t;
+    length : Uutil.Filesize.t }
+
 type t =
   { perm : Perm.t;
     uid : Uid.t;
@@ -673,6 +709,26 @@ type t =
     typeCreator : TypeCreator.t;
     length : Uutil.Filesize.t }
 
+let m = Umarshal.(prod6 Perm.m Uid.m Gid.m Time.m TypeCreator.m Uutil.Filesize.m
+                    (fun {perm; uid; gid; time; typeCreator; length} -> perm, uid, gid, time, typeCreator, length)
+                    (fun (perm, uid, gid, time, typeCreator, length) -> {perm; uid; gid; time; typeCreator; length}))
+
+let to_compat251 (p : t) : t251 =
+  { perm = p.perm;
+    uid = p.uid;
+    gid = p.gid;
+    time = p.time;
+    typeCreator = p.typeCreator;
+    length = p.length }
+
+let of_compat251 (p : t251) : t =
+  { perm = p.perm;
+    uid = p.uid;
+    gid = p.gid;
+    time = p.time;
+    typeCreator = p.typeCreator;
+    length = p.length }
+
 let template perm =
   { perm = perm; uid = Uid.dummy; gid = Gid.dummy;
     time = Time.dummy; typeCreator = TypeCreator.dummy;
@@ -687,6 +743,18 @@ let hash p h =
           (Time.hash p.time
              (TypeCreator.hash p.typeCreator h))))
 
+(* IMPORTANT!
+   This is the 2.51-compatible version of [hash]. It must always produce exactly
+   the same result as the [hash] in version 2.51.5.
+   If code changes elsewhere make this function produce a different result then
+   it must be updated accordingly to again return the 2.51-compatible result. *)
+let hash251 (p : t251) h =
+  Perm.hash p.perm
+    (Uid.hash p.uid
+       (Gid.hash p.gid
+          (Time.hash p.time
+             (TypeCreator.hash p.typeCreator h))))
+
 let similar p p' =
   Perm.similar p.perm p'.perm
     &&
@@ -803,6 +871,8 @@ let validatePrefs = Perm.validatePrefs
 
 type dirChangedStamp = Uutil.Filesize.t
 
+let mdirChangedStamp = Uutil.Filesize.m
+
 let freshDirStamp () =
   let t =
     (Unix.gettimeofday () +. sqrt 2. *. float (Unix.getpid ())) *. 1000.
